import re
import fsspec
import tempfile
import warnings
import logging

import pandas as pd

def make_partitions_df(glob = None, df = None, header = 'infer', sep = ','): 
    assert glob is not None or df is not None, "Either glob or df must be specified"
    df = pd.read_csv(f"filecache::{glob}", sep = sep, header = header) if glob else df
    assert set(df.columns) == set(['id', 'count']), "The metadata source must include id and count columns"
    df = df.sort_values(by='id', ignore_index=True)
    df['end_idx'] = df['count'].cumsum()
    df['start_idx'] = df['end_idx'] - df['count']
    df = df[['id', 'start_idx', 'end_idx']]
    return df

def objs_to_partitions_df(objs, header = None):
    rows = []
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        for idx, obj in enumerate(objs):
            try:
                df = pd.read_csv(f"filecache::{obj}", header = header)
                rows.append({"id": idx, "count": len(df),})
            except pd.errors.EmptyDataError:           
                pass
                # ignore empty shards      
                # rows.append({"id": idx, "count": 0,})
    return pd.DataFrame(rows)

def df_read_spec(df, row_count, row_start_idx, row_end_idx):

    assert row_start_idx > -1 and row_start_idx < row_count, f"row_start_idx must be in the range({0, row_count}]"
    assert row_end_idx > -1 and row_end_idx < row_count, f"row_end_idx must be in the range({0, row_count}]"

    start_obj_idx = df[ (row_start_idx >= df['start_idx']) & (row_start_idx < df['end_idx']) ]['id'].item()
    #max is needed to handle the case when the row_end_idx matches the end row of a partition and the start row of a partition
    #if the row_end_idx matches 2 partitions, calling item results in ValueError: can only convert an array of size 1 to a Python scalar
    end_obj_idx_res = df[ (row_end_idx >= df['start_idx']) & (row_end_idx <= df['end_idx']) ]['id'].max()    
    end_obj_idx = end_obj_idx_res if isinstance(end_obj_idx_res, int) else end_obj_idx_res.item()

    start_obj_idx = df [ df['id'] == start_obj_idx ].index.values.item()
    end_obj_idx = df [ df['id'] == end_obj_idx ].index.values.item()

    if row_end_idx > row_start_idx:
        
        df_spec = df.iloc[start_obj_idx : end_obj_idx + 1]
        
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            df_spec.loc[:, 'skiprows'] = df_spec['start_idx'].apply(lambda i: max(0, row_start_idx - i))
            df_spec.loc[:, 'numrows'] = df_spec['end_idx'].apply(lambda i: min(i, row_end_idx)) - (df_spec['start_idx'] + df_spec['skiprows'])
            
        df_values = df_spec[['skiprows', 'numrows', 'uri']].values
    
    #handle the wrap-around case
    #when the shard consists
    #of a tail of a dataset
    #followed by the dataset head
    else:

        head_df = df.iloc[start_obj_idx : ]
        tail_df = df.iloc[ : end_obj_idx + 1] #if end_obj_idx > start_obj_idx else None
        
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            head_df.loc[:, 'skiprows'] = head_df['start_idx'].apply(lambda i: max(0, row_start_idx - i))
            head_df.loc[:, 'numrows'] = head_df['end_idx'] - (head_df['start_idx'] + head_df['skiprows'])

            if tail_df is not None:
                tail_df.loc[:, 'skiprows'] = 0
                tail_df.loc[:, 'numrows'] = tail_df['end_idx'].apply(lambda i: min(i, row_end_idx)) - tail_df['start_idx']

        df_values = pd.concat([head_df, tail_df], ignore_index = True)[['skiprows', 'numrows', 'uri']].values
        
    return [ (skiprows, numrows, uri) for (skiprows, numrows, uri) in df_values ]

class BaseObjectStorageDataset():        
    def __init__(self, 
                 glob, 
                 header = 'infer', 

                 index_shards = None,
                 partitions_glob = None, 
                 make_partitions_df_fn = make_partitions_df, 
                 objs_to_partitions_df_fn = objs_to_partitions_df,

                 stats_glob = True,

                 worker = 0, 
                 replicas = 1, 
                 shard_size = None,   
                 batch_size = None,                
                 max_shards = None,

                 storage_options = None, 
                 cache_dir = None,):
                 
        assert glob, "glob must be specified"
        self.glob = glob

        self.header = header

        #init worker and replicas        
        assert worker is not None or replicas is not None, "Both worker and replicas must be specified together along with shard_size"
        assert worker > -1 and worker < replicas, f"worker must be in the range [0, {replicas})"
        self.worker = worker
        self.replicas = replicas
        if self.replicas > 1:
            assert shard_size, "shard_size must be specified if replicas is greater than 1"

        #resolve batch_size, shard_size, and max_shards
        if shard_size is not None and batch_size is not None:
            assert type(batch_size) is int and batch_size > 0 and type(shard_size) is int and shard_size > 0, "If batch_size and shard_size are specified, both must be greater than zero integers"
            #TODO: For now default to the smaller size
            if batch_size <= shard_size:
                logging.warn(f"defaulting to shard_size of {batch_size}")
                shard_size = batch_size
            else:
                logging.warn(f"defaulting to batch_size of {shard_size}")
                batch_size = shard_size
        elif shard_size is not None and batch_size is None:
            assert type(shard_size) is int and shard_size > 0, "The value for shard_size must be a greater than zero integer"
            logging.warn(f"defaulting to batch_size of {shard_size}")
            batch_size = shard_size
        elif batch_size is not None and shard_size is None:
            assert type(batch_size) is int and batch_size > 0, "The value for batch_size must be a greater than zero integer"
            shard_size = batch_size
        elif batch_size is None and shard_size is None:
            logging.warn(f"neither shard_size nor batch_size are specified, defaulting to batch_size=1 and shard_size=1")
            batch_size = shard_size = 1
        assert batch_size is not None or shard_size is not None, "either batch_size or shard_size must be specified"
        
        self.shard_size = shard_size
        self.max_shards = max_shards if max_shards else float('nan')

        # set the platform-specific temporary directory
        cache_dir = cache_dir if cache_dir else tempfile.gettempdir()

        # find out the protocol of the glob, e.g. s3, gs, hdfs, etc
        protocol, _ = fsspec.core.split_protocol(glob)
#         eager_load_batches = True if protocol in ('file') and eager_load_batches is None else eager_load_batches

        # use anonymous connection unless specified otherwise
        storage_options = storage_options if storage_options else {'anon': True}
        # handle the edge case when the user passes the boolean as a non boolean, e.g. str
        if 'anon' in storage_options:
            if type(storage_options['anon']) is str:
                storage_options['anon'] = True if str(storage_options['anon']).lower() == 'true' else False
            else:
                storage_options['anon'] = bool(storage_options['anon'])
        
        if protocol in ('http', 'https'):
            # setup a caching filesystem
            self.fs = fsspec.filesystem("filecache",
                                        target_protocol = protocol,
                                        target_options = storage_options,
                                        cache_storage = cache_dir)

        else:
            # setup a caching filesystem
            self.fs = fsspec.filesystem("filecache",
                                        target_protocol = protocol,
                                        target_options = storage_options,
                                        cache_storage = cache_dir)

        # get the object paths matching the glob
        if protocol in ('http', 'https'):
            _, uri = fsspec.core.split_protocol(glob)
            self.objs = [uri]
        else:
            if self.fs.isdir(glob):
                self.objs = self.fs.glob(f"{remove_trailing_forward_slash(glob)}/*")
            else:
                self.objs = self.fs.glob(glob)

        if not isinstance(self.objs, list) or not len(self.objs):
            raise RuntimeWarning(f"Specified glob pattern {self.glob} failed to match any objects. Ensure that the permission setting are configured correctly using storage_options before trying again.")

        self.objs = [f"{protocol}://{obj}" for obj in self.objs]

        #create a stats dataframe unless this feature is turned off
        if stats_glob is None or stats_glob is False:
            self.stats_df = None

        else:
            #if not specified explicitly try to fall back to default location
            if type(stats_glob) is not str and protocol not in ('http', 'https'):

                #if the glob is a file or a wildcard
                if self.fs.isfile(glob) or (not self.fs.isfile(glob) and not self.fs.isdir(glob)):
                    maybe_stats_glob = f"{protocol}://{self.fs._parent(glob)}/.meta/stats/"
                elif self.fs.isdir(glob):
                    #remove trailing / characters
                    maybe_stats_glob = f"{remove_trailing_forward_slash(glob)}/.meta/stats/"                
                else:
                    logging.warn(f"failed to construct a stats dataframe path for glob={glob}")
                    maybe_stats_glob = None

                maybe_stats = self.fs.glob(maybe_stats_glob)
                if not isinstance(maybe_stats, list) or not len(maybe_stats):
                    logging.warn(f"stats_glob is not explicitly specified at initialization, failed to find stats metadata at default location {maybe_stats_glob}")
                    stats_glob = None
                else:
                    [stats_glob] = maybe_stats
                    stats_glob = f"{protocol}://{stats_glob}"
                    logging.warn(f"stats_glob is not specified at initialization, defaulting to stats_glob={stats_glob}")            

            else:
                stats_glob = None
            
            if stats_glob is not None:
                self.stats_df = pd.read_csv(f"filecache::{stats_glob}")

        #figure out the metadata
        if index_shards is None:   
            
            #ignore fall back with protocols that dont support glob
            if partitions_glob is None and protocol not in ('http', 'https'):

                #if the glob is a file or a wildcard
                if self.fs.isfile(glob) or (not self.fs.isfile(glob) and not self.fs.isdir(glob)):
                    maybe_partitions_glob = f"{protocol}://{self.fs._parent(glob)}/.meta/shards/"
                elif self.fs.isdir(glob):
                    #remove trailing / characters
                    maybe_partitions_glob = f"{remove_trailing_forward_slash(glob)}/.meta/shards/"                
                else:
                    logging.warn(f"failed to construct a shards dataframe path for glob={glob}")
                    maybe_partitions_glob = None
                
                maybe_shards = self.fs.glob(maybe_partitions_glob)
                if not isinstance(maybe_shards, list) or not len(maybe_shards):
                    logging.warn(f"partitions_glob is not specified at initialization, failed to find shard metadata at default location {maybe_partitions_glob}")
                else:
                    [partitions_glob] = maybe_shards
                    partitions_glob = f"{protocol}://{partitions_glob}"
                    logging.warn(f"partitions_glob is not specified at initialization, defaulting to partitions_glob={partitions_glob}")

            if partitions_glob is None:
                index_shards = True
                logging.warn("partitions_glob is not specified at initialization, attempting to proceed with index_shards=True which can take a while for large objects.")
            else:
                index_shards = False
        else:
            assert isinstance(index_shards, bool), "index_shards must be set to True or False"

        assert make_partitions_df_fn, "make_partitions_df_fn must be specified"
        self.make_partitions_df_fn = make_partitions_df_fn

        if index_shards is None \
                or index_shards is True \
                or partitions_glob is not None:
            assert objs_to_partitions_df_fn, "objs_to_partitions_df_fn must be specified"
            self.objs_to_partitions_df_fn = objs_to_partitions_df_fn


        # if the shard metadata is unavailable
        # use the objects in the glob to generate the metadata
        # with the count of lines of records per shard            
        if index_shards is None or index_shards is True:            
            self.partitions_df = make_partitions_df_fn(df = objs_to_partitions_df_fn(self.objs, header = self.header))
        else:            
            self.partitions_df = make_partitions_df_fn(glob = partitions_glob)
                
        # if the number of shards is identical to the number of metadata records
        if len(self.objs) == len(self.partitions_df):
            self.partitions_df = self.partitions_df.sort_values(by = 'id', ascending = True)
            self.partitions_df['uri'] = pd.Series(name = 'uri', data = self.objs).sort_values(ascending=True)

        else:            
            # query for obj sizes in order to later ignore the empty shards 
            # because they are not in the metadata dataframe
            if self.fs.isdir(glob):
                glob_dict = self.fs.glob(f"{remove_trailing_forward_slash(glob)}/*", detail = True)
            else:
                glob_dict = self.fs.glob(glob, detail = True)

            kvs = glob_dict.items()

            objs_df = pd.DataFrame(data = map( lambda kv : (kv[1]['Key'], kv[1]['Size']), kvs ), columns = ['key', 'size'] )
            # use just the non-empty shards / partitions
            objs_df = objs_df[ objs_df['size'] > 0 ]

            assert len(objs_df) == len(self.partitions_df), f"The number of the non-empty objects matching the glob pattern {len(objs_df)} is not equal to the number of the non-empty shards {len(self.partitions_df)}."

            self.partitions_df['uri'] = objs_df['key'].apply(lambda s: f"{protocol}://{s}").sort_values(ascending=True)

        #TODO: improve: right now assumes that the uri has the matching ID specified in 'part-00000-' substring inthe file
        # objs_df = pd.DataFrame(columns=['uri'], data = pd.Series(self.objs).apply(lambda s: f"{protocol}://{s}"))
        # id_re = re.compile('part-(\d{5})-')
        # objs_df['id'] = objs_df['uri'].apply(lambda s: int(id_re.findall(s)[0]))
        # self.partitions_df = self.partitions_df.merge(objs_df, on = 'id')
        #TODO: improve

        self.obj_count = len(self.partitions_df)
        self.row_count = self.partitions_df.iloc[-1]['end_idx']

    def __len__(self):
        if self.partitions_df is not None:
            return self.partitions_df.iloc[-1]['end_idx'] + 1
        else:
            return -1

    def __iter__(self):
        shard_start_idx = self.worker * self.shard_size
        while self.max_shards:
            try:
                num_rows, row_start_idx, row_end_idx = self.shard_size, shard_start_idx, shard_start_idx + self.shard_size

                # the for loop handles the case where the shard
                # is larger than the number of rows in the dataset
                row_start_indicies = list(range(row_start_idx, row_end_idx, self.row_count))
                row_end_indicies = list(row_start_indicies[1:] + [row_end_idx])
                src_df_list = []
                for row_start_idx, row_end_idx in zip(row_start_indicies, row_end_indicies):

                    #query rows based on actual indicies in the dataset
                    row_start_idx = row_start_idx % self.row_count
                    row_end_idx = row_end_idx % self.row_count

                    if self.header is None:
                        for (skiprows, nrows, uri) in df_read_spec(self.partitions_df, self.row_count, row_start_idx, row_end_idx):
                            src_df_list.append( pd.read_csv(f"filecache::{uri}", header = None, nrows=nrows, skiprows=skiprows) )

                    elif self.header == 'infer' or self.header == 0:                        
                        for (skiprows, nrows, uri) in df_read_spec(self.partitions_df, self.row_count, row_start_idx, row_end_idx):
                            
                            if (skiprows == 0):
                                #skiprows == 0 when we are at the top of the file with a (potential) header                            
                                df = pd.read_csv(f"filecache::{uri}", header = self.header, nrows = nrows, skiprows = 0)

                                # set columns to numbers since reseting the index drops the data types
                                df.columns = list(range(len(df.columns)))

                                #silly hack because pandas doesn't let me to reset the index along the columns axis
                                # df = df.T
                                # df.reset_index(drop = True, inplace = True)
                                # df = df.T

                            else:
                                #(skiprows+1) skip the extra row to account for the skipped header
                                df = pd.read_csv(f"filecache::{uri}", header = None, nrows = nrows, skiprows = skiprows + 1) 

                            # empty data frames don't mess with the data
                            # but mess with the schema so ignore them
                            if (len(df)) > 0:
                                src_df_list.append(df)

                    else:
                        raise ValueError(f"header attribute cannot be set to {self.header}")

                    # map(lambda df: df.columns, src_df_list)
                    # assert len(set(map(lambda df: df.shape[1], src_df_list))) == 1, f"Failed to process in the following fragment of the dataset {df_read_spec(self.partitions_df, self.row_count, row_start_idx, row_end_idx)} due to a change in the number of the columns."

                yield pd.concat(
                    objs = src_df_list,
                    copy = False,
                    ignore_index=True,
                )
                                
            finally:
                shard_start_idx += (self.replicas * self.shard_size)
                self.max_shards -= 1

def remove_trailing_forward_slash(s):
    return re.split('/+$', s)[0]


import os
import click
import uuid

from dotenv import dotenv_values

VERSION = "0.0.1"
IMAGE_NAME = "kaenai/kaen"
IMAGE_TAG = "latest"
IMAGE = f"{IMAGE_NAME}:{IMAGE_TAG}"

def get_backend_provider(dojo = None, job = None):
	if job:
		job_config = {
			**dotenv_values(f"{os.getcwd()}/.job/.{job}/env.sh")
		}
		dojo_config = {
			**dotenv_values(f"{os.getcwd()}/.dojo/.{job_config['KAEN_DOJO']}/env.sh")
		}
		backend = dojo_config['KAEN_BACKEND']
		provider = dojo_config['KAEN_PROVIDER']
		return backend, provider		
	elif dojo:
		dojo_config = {
			**dotenv_values(f"{os.getcwd()}/.dojo/.{dojo}/env.sh")
		}
		backend = dojo_config['KAEN_BACKEND']
		provider = dojo_config['KAEN_PROVIDER']
		return backend, provider
	else:
		return None

def get_volumes(provider = None):
	volumes = {f"{os.getcwd()}" : {"bind": "/workspace", "mode": "rw"},}
	if provider == 'local':
		volumes = {
			**volumes, 
			**{"/var/run/docker.sock": {"bind": "/var/run/docker.sock", "mode": "rw"}}
		}
	return volumes

def assert_aws_credentials():
	assert 'AWS_ACCESS_KEY_ID' in os.environ, "Missing environment variable AWS_ACCESS_KEY_ID"
	assert 'AWS_DEFAULT_REGION' in os.environ, "Missing environment variable AWS_DEFAULT_REGION"
	assert 'AWS_SECRET_ACCESS_KEY' in os.environ, "Missing environment variable AWS_SECRET_ACCESS_KEY"

def prepare_docker():
	prepare_docker_image(f"{IMAGE_NAME}:{IMAGE_TAG}", IMAGE_NAME, IMAGE_TAG)

def prepare_docker_image(image, name, tag):
	import docker
	client = docker.from_env()
	images = client.images.list(image)
	if len(images):
		image = images[0]
	else:
		print(f'Unable to find {image}:{tag} in local storage. Pulling the image...')
		image = client.images.pull(name, tag)
	return client

@click.group()
def cli():
	f"""Kaen helps you train distributed PyTorch deep learning models in AWS, Azure, and Google Cloud.

	You are using Kaen version {VERSION}.
	"""
	pass

@cli.command("init")
@click.option('--provider', required=True, type=click.Choice(['aws', 'azure', 'gcp', 'local'], case_sensitive=False))
@click.option('--backend', default='docker', type=click.Choice(['docker', 'k8s'], case_sensitive=False), show_default=True)
@click.option('--worker-instances', default='1', type=int, show_default=True)
@click.option('--worker-instance-type', default='t3.micro', show_default=True)
@click.option('--manager-instances', default='1', type=int, show_default=True)
@click.option('--manager-instance-type', default='t3.micro', show_default=True)
@click.option('--volume-size', default='8', type=int, show_default=True)
@click.option('--ssh-connection-timeout', default='5', type=int, show_default=True)
@click.option('--ssh-connection-attempts', default='1', type=int, show_default=True)
def init(provider, 
			backend, 
			worker_instances, 
			worker_instance_type, 
			manager_instances, 
			manager_instance_type,
			volume_size,
			ssh_connection_timeout,
			ssh_connection_attempts):			
	"""Initialize a training dojo in a specified infrastructure provider."""
	DojoCLI.init(provider, 
			backend, 
			worker_instances, 
			worker_instance_type, 
			manager_instances, 
			manager_instance_type,
			volume_size,
			ssh_connection_timeout,
			ssh_connection_attempts)

class DojoCLI(click.MultiCommand):
	@cli.command("init")
	@click.option('--provider', required=True, type=click.Choice(['aws', 'azure', 'gcp', 'local'], case_sensitive=False))
	@click.option('--backend', default='docker', type=click.Choice(['docker', 'k8s'], case_sensitive=False), show_default=True)
	@click.option('--worker-instances', default='1', type=int, show_default=True)
	@click.option('--worker-instance-type', default='t3.micro', show_default=True)
	@click.option('--manager-instances', default='1', type=int, show_default=True)
	@click.option('--manager-instance-type', default='t3.micro', show_default=True)
	@click.option('--volume-size', default='8', type=int, show_default=True)
	@click.option('--ssh-connection-timeout', default='5', type=int, show_default=True)
	@click.option('--ssh-connection-attempts', default='1', type=int, show_default=True)
	def init(provider, 
				backend, 
				worker_instances, 
				worker_instance_type, 
				manager_instances, 
				manager_instance_type,
				volume_size,
				ssh_connection_timeout,
				ssh_connection_attempts):			
		"""Initialize a training dojo in a specified infrastructure provider."""
		assert provider == 'aws	' or provider == 'local', f"Provider {provider} is not publicly available in version {VERSION}"
		assert backend == 'docker', f"Backend {backend} is not publicly available in version {VERSION}"

		#check for the aws settings
		if (provider == 'aws'):
			assert_aws_credentials()

			# #drop the provider prefix for the instance type, e.g. aws:t3.micro -> t3.micro
			# assert worker_instance_type[:len(provider) + 1] == provider + ":",\
			# 	f"Worker instance type {worker_instance_type} is missing a prefix {provider}:"
			# worker_instance_type = worker_instance_type[len(provider) + 1:]

			# manager_instance_type = manager_instance_type[len(provider) + 1:]

		print(f"Checking for the latest updates...", end='')
		client = prepare_docker()
		print('done')

		dojo = os.environ['KAEN_DOJO'] if 'KAEN_DOJO' in os.environ else None
		while not dojo or dojo[:1].isalpha() is False:
			dojo = uuid.uuid4().hex[:8]
		print(f"Using a dojo store at {os.getcwd()}/.dojo/.{dojo}")

		environment = { "KAEN_DOJO": dojo, 
						"KAEN_VERSION": VERSION,
						"KAEN_BACKEND": backend,
						"KAEN_PROVIDER": provider,
						}

		volumes = {f"{os.getcwd()}" : {"bind": "/workspace", "mode": "rw"},}
		
		if provider == 'local':
			print(f"Preparing a dojo using {backend} on {provider} ...")
			volumes = {
				**volumes, 
				**{"/var/run/docker.sock": {"bind": "/var/run/docker.sock", "mode": "rw"}}
			}
		else:
			environment = { **environment, **{
							"KAEN_DOJO_CONNECT_TIMEOUT": ssh_connection_timeout,
							"KAEN_DOJO_CONNECT_ATTEMPTS": ssh_connection_attempts,
							"AWS_ACCESS_KEY_ID": os.environ['AWS_ACCESS_KEY_ID'],
							"AWS_DEFAULT_REGION": os.environ['AWS_DEFAULT_REGION'],
							"AWS_SECRET_ACCESS_KEY": os.environ['AWS_SECRET_ACCESS_KEY'],
							"TF_VAR_instances": worker_instances,
							"TF_VAR_worker_instance_type": worker_instance_type,
							"TF_VAR_manager_instances": manager_instances,
							"TF_VAR_manager_instance_type": manager_instance_type,
							"TF_VAR_volume_size": volume_size,}
			}		
			print(f"Preparing a dojo using {backend} on {provider} with {worker_instances} worker(s) and {manager_instances} manager(s) ...")

		container = client.containers.run(	image = IMAGE, 
			command=f"/bin/bash -c '/opt/kaen/{backend}/{provider}/dojo_init.sh'",
			volumes = volumes,
			environment = environment,
			auto_remove = True,
			detach=True)
			
		logs = container.attach(stream = True, logs = True)
		for s in logs:
			print(s.decode('utf-8').strip())	

		print(dojo)

	@click.command()	
	@click.argument('dojo')
	def activate(dojo):
		print(f"Activating dojo {dojo}")
		backend, provider = get_backend_provider(dojo)
		if 'aws' == provider:
			assert_aws_credentials()

		client = prepare_docker()
		volumes = get_volumes(provider)

		container = client.containers.run(	
			image = IMAGE, 
			command = f"/bin/bash -c '/opt/kaen/{backend}/{provider}/dojo_activate.sh'",
			volumes = volumes ,
			environment = { "KAEN_DOJO": dojo, 
							},
			auto_remove = True,
			detach = True )

		logs = container.attach(stream = True, logs = True)
		for s in logs:
			print(s.decode('utf-8'))	

	@click.command()	
	@click.option('--reverse', '-r',  is_flag = True, default = False)
	def ls(reverse):		
		"""Output a list with the training dojo(s) in the current workspace.		
		By default, the dojo that was created first in the workspace is listed first."""

		client = prepare_docker()
		
		environment = {
			"KAEN_DOJO_LS_REVERSE": "" if reverse else "-r"
		}

		container = client.containers.run(	
			image = IMAGE, 
			command=f"/bin/bash -c '/opt/kaen/dojo_ls.sh'",
			volumes = get_volumes() ,
			environment = environment,
			auto_remove = True,
			detach = True )

		logs = container.attach(stream = True, logs = True)
		for s in logs:
			print(s.decode('utf-8').strip())	

	@click.command()
	def prune():
		client = prepare_docker()
		container = client.containers.run(	
			image = IMAGE, 
			command=f"/bin/bash -c '/opt/kaen/dojo_prune.sh'",
			volumes = get_volumes() ,
			auto_remove = True,
			detach = True )

		logs = container.attach(stream = True, logs = True)
		for s in logs:
			s = s.decode('utf-8').strip()
			if len(s):
				print(s)
	
	@click.command()
	@click.argument('dojo')
	def rm(dojo):
		print(f"Removing dojo {dojo} and deactivating store {os.getcwd()}/.dojo/.{dojo}")

		backend, provider = get_backend_provider(dojo)
		environment = {}
		if 'aws' == provider:
			assert_aws_credentials()
			environment = {
				"AWS_ACCESS_KEY_ID": os.environ['AWS_ACCESS_KEY_ID'],
				"AWS_DEFAULT_REGION": os.environ['AWS_DEFAULT_REGION'],
				"AWS_SECRET_ACCESS_KEY": os.environ['AWS_SECRET_ACCESS_KEY'],				
			}
		
		client = prepare_docker()

		volumes = get_volumes(provider)
		
		container = client.containers.run(	
			image = IMAGE, 
			command=f"/bin/bash -c '/opt/kaen/{backend}/{provider}/dojo_rm.sh'",
			volumes = volumes,
			environment = { **{"KAEN_DOJO": dojo,}, 
							**environment,},
			auto_remove = True,
			detach=True )

		logs = container.attach(stream = True, logs = True)
		for s in logs:
			print(s.decode('utf-8'))	
	
	@click.command()	
	@click.argument('dojo')
	def inspect(dojo):	
		"""Return key value settings for the specified dojo."""
		backend, provider = get_backend_provider(dojo)
		if 'aws' == provider:
			assert_aws_credentials()

		client = prepare_docker()
		
		container = client.containers.run(	
			image = IMAGE, 
			command=f"/bin/bash -c '/opt/kaen/dojo_inspect.sh'",
			volumes = get_volumes(provider) ,
			environment = { "KAEN_DOJO": dojo, 
							},
			auto_remove = True,
			detach = True )

		logs = container.attach(stream = True, logs = True)
		for s in logs:
			s = s.decode('utf-8').strip()
			if len(s):
				print(s)

	def list_commands(self, ctx):
		return ['ls', 'rm', 'prune', 'activate', 'inspect', 'init']

	def get_command(self, ctx, name):
		return {"ls": DojoCLI.ls,
				"rm": DojoCLI.rm,
				"prune": DojoCLI.prune,
				"inspect": DojoCLI.inspect,
				"init": DojoCLI.init,
				'activate': DojoCLI.activate}[name]

@cli.command(cls=DojoCLI)
def dojo():
	"""Manage a dojo training environment."""
	pass

class JobCLI(click.MultiCommand):
	@click.command()
	@click.option('--dojo', required = True)
	@click.option('--image', required = True)
	@click.option('--subnet', required = False)
	@click.option('--manager-ip', required = False)
	def create(image, dojo, subnet, manager_ip):
		"""Create a training job in a specified dojo training environment. 
		The training job must be packaged as a container and available to docker pull 
		command, in other words docker pull <image> must pull the image."""

		if subnet is not None or manager_ip is not None:
			assert subnet is not None and manager_ip is not None, "Both --subnet and --manager-ip must be specified"
			from ipaddress import ip_network, ip_address
			net = ip_network(subnet)
			assert ip_address(manager_ip) in net, f"Manager IP {manager_ip} is not in the CIDR subnet {subnet} range"

		job = None
		while not job or job[:1].isalpha() is False:
			job = uuid.uuid4().hex[:8]			
		print(f"Creating a job {job} using image {image} in dojo {dojo}")

		backend, provider = get_backend_provider(dojo)

		print(f"Checking for the latest version...", end='')
		client = prepare_docker()
		print('done')

		container = client.containers.run(image = IMAGE, 
			command=f"/bin/bash -c '/opt/kaen/{backend}/{provider}/job_create.sh'",
			volumes = get_volumes(provider) ,
			environment = { "KAEN_DOJO": dojo,
							"KAEN_JOB": job,
							"KAEN_JOB_IMAGE": image,
							"KAEN_JOB_SUBNET": subnet if subnet else None,
							"KAEN_JOB_MANAGER_IP": manager_ip if subnet else None,
							},
			auto_remove = True,
			detach=True)
		
		logs = container.attach(stream = True, logs = True)
		for s in logs:
			print(s.decode('utf-8').strip())	
		
		print(job)			

	@click.command()	
	@click.argument('job')
	def inspect(job):
		"""Return key value settings for the specified job."""
		
		client = prepare_docker()
		
		container = client.containers.run(	
			image=IMAGE, 
			command=f"/bin/bash -c '/opt/kaen/{backend}/{provider}/job_inspect.sh'",
			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
											"mode": "rw"}} ,
			environment = { "KAEN_JOB": job, },
			auto_remove = False,
			detach=True )

		logs = container.attach(stream = True, logs = True)
		for s in logs:
			print(s.decode('utf-8'))	

	@click.command()
	@click.argument('job')
	@click.option('--replicas', default='1', type=int)
	@click.option('--detach', is_flag = True, default = False)
	@click.option('--port', '-p', required = False, default = [], multiple = True, type=click.Tuple([str, str]) )	
	@click.option('--env', '-e', required = False, default = [], multiple = True, type=click.Tuple([str, str]) )
	def start(job, replicas, detach, port, env):
		"""Start an existing job using the specified number of worker replicas."""

		print(f"Checking for the latest version...", end='')
		client = prepare_docker()
		print('done')

		backend, provider = get_backend_provider(job = job)		

		env = dict({ "KAEN_JOB": job,
					 "KAEN_JOB_REPLICAS": replicas,
					 "KAEN_JOB_CFG_DETACH": 1 if detach else 0,
					 },
					 **{"KAEN_JOB_ARG_" + k : v for k, v in env},
					 **{"KAEN_JOB_PORT_" + k : v for k, v in port})

		vol = get_volumes(provider)

		container = client.containers.run(	image = IMAGE, 
			command=f"/bin/bash -c '/opt/kaen/{backend}/{provider}/job_start.sh'",
			volumes = vol ,
			environment = env,
			auto_remove = True,
			detach = True)

		if not detach:
			logs = container.attach(stream = True, logs = True)
			for s in logs:
				print(s.decode('utf-8').strip())	

	@click.command()	
	@click.option('--reverse', '-r',  is_flag = True, default = False)
	def ls(reverse):		
		"""List existing jobs in the current workspace.
		By default, the dojo that was created first in the workspace is listed first."""

		client = prepare_docker()
		
		env = {
			"KAEN_JOB_LS_REVERSE": "" if reverse else "-r"
		}
		
		container = client.containers.run(	
			image=IMAGE, 
			command=f"/bin/bash -c '/opt/kaen/job_ls.sh'",
			environment = env,
			volumes = get_volumes(),
			auto_remove = False,
			detach = True )

		logs = container.attach(stream = True, logs = True)
		for s in logs:
			print(s.decode('utf-8').strip())

	@click.command()
	@click.argument('job')
	def rm(job):
		"""Remove the specified job and associated resources."""
		client = prepare_docker()
		env = dict({ "KAEN_JOB": job,})

		container = client.containers.run(	
			image=IMAGE, 
			command=f"/bin/bash -c '/opt/kaen/{backend}/{provider}/hpo_disable.sh'",
			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
											"mode": "rw"}} ,
			environment = env,
			auto_remove = False,
			detach = True )

		logs = container.attach(stream = True, logs = True)
		for s in logs:
			print(s.decode('utf-8').strip())			
		
		container = client.containers.run(	
			image=IMAGE, 
			command=f"/bin/bash -c '/opt/kaen/{backend}/{provider}/job_rm.sh'",
			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
											"mode": "rw"}} ,
			environment = env,
			auto_remove = False,
			detach = True )

		logs = container.attach(stream = True, logs = True)
		for s in logs:
			print(s.decode('utf-8').strip())	


	def list_commands(self, ctx):
		return ['ls', 'rm', 'create', 'inspect', 'start']

	def get_command(self, ctx, name):
		return {"ls": JobCLI.ls,
				"rm": JobCLI.rm,
				'create': JobCLI.create,
				'inspect': JobCLI.inspect,
				"start": JobCLI.start}[name]

@cli.command("job", cls=JobCLI)
def job():
	"""Manage jobs in a specific dojo training environment."""
	pass

class HpoCLI(click.MultiCommand):
	def list_commands(self, ctx):
		return ["enable", "disable", 'start', 'stop', 'inspect', 'set']

	def get_command(self, ctx, name):
		return {"enable": HpoCLI.enable,
				"disable": HpoCLI.disable,
				"start": HpoCLI.start,
				"stop": HpoCLI.stop,
				"client": HpoCLI.client,
				'inspect': HpoCLI.inspect,
				'set': HpoCLI._set,
				'kv': HpoCLI.kv}[name]

	@click.command()
	@click.argument('name')
	@click.option('--parent-run', is_flag = True, required = False)
	@click.option('--active-run-params', is_flag = True, required = False)
	@click.option('--hp-set-id', required = False)
	def inspect(name, parent_run, active_run_params, hp_set_id):
		from kaen.hpo.mlflow_utils import get_client, get_parent_run, \
			get_active_child_run, get_active_run_params, \
			get_active_run_hp_set_id

		mlflow_tracking_uri = os.environ['MLFLOW_TRACKING_URI'] if 'MLFLOW_TRACKING_URI' in os.environ else None
		client = get_client(mlflow_tracking_uri)
		if parent_run:
			run = get_parent_run(client, name)
			if run is not None:
				for k, v in {"EXPERIMENT_RUN_ID": run.info.run_id, 
							"EXPERIMENT_RUN_STATUS": run.info.status, 
							# **run.data.tags, 
							}.items():
					print(f"KAEN_{k}={v}")
			else:
				return None
		elif active_run_params:
			params = get_active_run_params(mlflow_tracking_uri = mlflow_tracking_uri, mlflow_experiment_name = name)
			if params is not None:
				print(params)
		elif hp_set_id:
			hp_id = get_active_run_hp_set_id(mlflow_experiment_name=name, hp_set_id=hp_set_id)
			print(hp_id)

	@click.command()
	@click.argument('job')
	@click.option('--image', required = True)
	@click.option('--num-runs', required = False, default = 1, show_default=True)
	@click.option('--seed', required = False, default = 42, show_default=True)
	@click.option('--service-prefix', required = False, default = 'kaen.hpo.services', show_default=True)
	@click.option('--service-name', required = False, default = 'BaseMLFlowService', show_default=True)
	@click.option('--port', '-p', required = False, default = [], multiple = True, type=click.Tuple([str, str]) )	
	@click.option('--env', '-e', required = False, default = [], multiple = True, type=click.Tuple([str, str]) )
	def enable(job, image, num_runs, seed, service_prefix, service_name, port, env):
		"""Enable hyperparameter optimization(HPO) for a job using
		the HPO service packaged in the specified image."""

		print(f"Checking for the latest version...", end='')
		client = prepare_docker()
		print('done')

		backend, provider = get_backend_provider(job = job)

		env = dict({ "KAEN_JOB": job,
					 "KAEN_HPO_IMAGE": image,
					 "KAEN_HPO_JOB_RUNS": num_runs,
					 "KAEN_HPO_SEED": seed,
					 "KAEN_HPO_SERVICE_PREFIX": service_prefix,
					 "KAEN_HPO_SERVICE_NAME": service_name,
					 },
					 **{"KAEN_JOB_ARG_" + k : v for k, v in env},
					 **{"KAEN_JOB_PORT_" + k : v for k, v in port})

		vol = get_volumes(provider)

		container = client.containers.run(	image = IMAGE, 
			command=f"/bin/bash -c '/opt/kaen/{backend}/{provider}/hpo_enable.sh'",
			volumes = vol,
			environment = env,
			auto_remove = True,
			detach = True)

		logs = container.attach(stream = True, logs = True)
		for s in logs:
			print(s.decode('utf-8').strip())	

	@click.command()
	@click.argument('job')
	def disable(job):
		"""Disable hyperparameter optimization(HPO) for a job.
		If a job never had HPO enabled, this command succeeds."""

		print(f"Checking for the latest version...", end='')
		client = prepare_docker()
		print('done')

		backend, provider = get_backend_provider(job = job)

		env = dict({ "KAEN_JOB": job,})
		vol = get_volumes(provider)

		container = client.containers.run(	image = IMAGE, 
			command=f"/bin/bash -c '/opt/kaen/{backend}/{provider}/hpo_disable.sh'",
			volumes = vol,
			environment = env,
			auto_remove = True,
			detach = True)

		logs = container.attach(stream = True, logs = True)
		for s in logs:
			print(s.decode('utf-8').strip())	

	@click.command()
	@click.argument('name')
	def _set(name):
		from kaen.hpo.mlflow_utils import get_active_run_feature_set_id		
		print(get_active_run_feature_set_id(feature_set_id = name))

	@click.command()
	@click.option('--output', default='bash', required = False)
	def kv(output):
		from kaen.hpo.mlflow_utils import get_active_run_params
		print(get_active_run_params(output = output))

	@click.command()
	@click.option('--num_runs', required = False, default = 1)
	@click.option('--client_prefix', required = False, default = 'kaen.hpo.client')
	@click.option('--client_name', required = False, default = 'BaseMLFlowClient')
	def client(num_runs, client_prefix, client_name):
		"""Execute a run(s) of a hyperparameter optimization service client.
		
		Unless specified otherwise, the client executes a single run. The default
		implementation of the client uses kaen.hpo.client.BaseMLFlowClient.
		This implementation simply queries for a seed hyperparameter and reports
		uniform random values in a range [0., 1.) as the values for the loss metric.
		"""
		import importlib.util
		mod = importlib.import_module(client_prefix)
		klass = getattr(mod, client_name)
		client = klass()
		client.run(num_runs = num_runs)

	@click.command()
	@click.option('--name', required = True)
	@click.option('--num_runs', required = False, default = 1)
	@click.option('--hpo_seed', required = False, default = 42)
	@click.option('--service_prefix', required = False, default = 'kaen.hpo.services')
	@click.option('--service_name', required = False, default = 'BaseMLFlowService')
	def start(name, num_runs, hpo_seed, service_prefix, service_name):
		import importlib.util
		mod = importlib.import_module(service_prefix)
		klass = getattr(mod, service_name)
		hpo_service = klass()
		hpo_service.run(mlflow_experiment_name = name,
						num_trials = num_runs, 
						hpo_seed = hpo_seed)

	@click.command()
	def stop():
		from kaen.hpo.mlflow_utils import terminate_parent_run
		terminate_parent_run()


@cli.command(cls=HpoCLI)
def hpo():
	"""Manage hyperparameter optimization."""
	pass

@cli.command("jupyter")
def jupyter():
	"""Work with a Jupyter notebook environment."""
	IdeCLI.jupyter()

class IdeCLI(click.MultiCommand):	
	def next_available_port(port=8888):
		import socket
		skt = None
		try:		
			skt = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			while port < 65536:
				try:
					skt.bind(('', port))
					return port
				except OSError as e:
					#port already bound
					if e.errno == 48:
						port += 1	
					else:
						raise e
		finally:
			if skt is not None:
				skt.close()
		return None

	@cli.command()
	def jupyter():
		"""Work with a Jupyter notebook environment."""
		image_name = "kaenai/jupyter"
		image_tag = "latest"
		image = f"{image_name}:{image_tag}"
		import time		
		from hashlib import md5
		token = md5(str.encode(str(time.time()))).hexdigest()
		port = IdeCLI.next_available_port()
		
		container = IdeCLI._run(image_name, image_tag, [('8888', port)], [('JPY_TKN', token)])
		logs = container.attach(stream = True, logs = True)
		for s in logs:
			s = s.decode('utf-8').strip()
			if "Use Control-C to stop this server" in s:
				break
			else:
				print(s)
		print(f"Started Jupyter. Attempting to navigate to Jupyter in your browser using http://127.0.0.1:{port}/?token={token}")
		import webbrowser
		webbrowser.open_new(f"http://127.0.0.1:{port}/?token={token}")
	
	def _run(image_name, image_tag, port = [], env = []):
		image = f"{image_name}:{image_tag}"
		client = prepare_docker_image(image = image, name = image_name, tag = image_tag)

		ports = {
			str(host): str(guest) for host, guest in port
		}
		environment = {
			str(k): str(v) for k, v in env
		}

		container = client.containers.run(image = image, 
			environment = environment,
			ports = ports,
			privileged = True,
			auto_remove = True,
			detach = True)

		return container

	@click.command()
	@click.option('--port', '-p', required = False, default = [], multiple = True, type=click.Tuple([str, str]) )	
	@click.option('--env', '-e', required = False, default = [], multiple = True, type=click.Tuple([str, str]) )
	def run(image_name, image_tag, port, env):
		"""Run an IDE image"""
		pass

	def list_commands(self, ctx):
		return [
			# 'jupyter', 
		# 'run',
		]

	def get_command(self, ctx, name):
		return {
			# "jupyter": IdeCLI.jupyter, 
		# "run": IdeCLI.run
		}[name]


@cli.command(cls=IdeCLI)
def ide():
	"""Manage an Integrated Development Environment(IDE)"""
	pass